import os
from typing import Callable, List

from django.http import HttpResponse

from djavue.renderers import VueRenderer, get_vue_template
from djavue.renderers.vuetify import VuetifyRenderer


class VueTemplate:
    """
    Shortcut for creating a view that returns a Vue file
    """

    renderer_class: VueRenderer = VueRenderer

    def _validate(self, request) -> None:
        """
        Checks if template_name and context exists
        """

        if self.context is None:
            raise Exception(f"{self.get_template_name(request)} context is null.")

    def get_context(self, request) -> object:
        """
        Function for defining context that must be overriden by the user
        """

        return None

    def get_extra_style_sheets(self, request) -> List[str]:
        """
        Function for defining aditional style_sheets to be loaded that can be overriden by the user
        """

        return []

    def get_extra_scripts(self, request) -> List[str]:
        """
        Function for defining aditional style_sheets to be loaded that can be overriden by the user
        """

        return []

    def get_template_name(self, request) -> str:
        """
        Function for programatically defining the template that can be overriden by the user
        """

        return self.Meta.template_name

    def _mount(self, request) -> HttpResponse:
        """
        Validates, loads the template and returns the content in Http
        """

        self.context: object = self.get_context(request)
        self._validate(request)

        style_sheets: List[str] = self.get_extra_style_sheets(request)
        scripts: List[str] = self.get_extra_scripts(request)

        instance: VueRenderer = get_vue_template(
            os.path.split(self.get_template_name(request))[1],
            title=self.Meta.page_title,
            renderer=self.renderer_class,
        )

        self.content: str = instance.render(self.context, style_sheets, scripts)

        return HttpResponse(self.content)

    @classmethod
    def as_view(cls) -> Callable:
        """
        Returns a function to be used in Django's path() or url()
        """

        instance = cls()
        return lambda request: instance._mount(request)

    class Meta:
        """
        Class that must be overriden by the user with the view's details
        """

        page_title = "Djavue Page"
        root_path = ""
        template_name = None
        context = None


class VuetifyTemplate(VueTemplate):
    renderer_class: VueRenderer = VuetifyRenderer

    def get_extra_style_sheets(self, request) -> List[str]:
        return [
            "https://fonts.googleapis.com/css?family=Roboto:100,300,400,500,700,900",
            "https://cdn.jsdelivr.net/npm/@mdi/font@6.x/css/materialdesignicons.min.css",
            "https://cdn.jsdelivr.net/npm/vuetify@2.x/dist/vuetify.min.css",
        ]

    def get_extra_scripts(self, request) -> List[str]:
        return ["https://cdn.jsdelivr.net/npm/vuetify@2.x/dist/vuetify.js"]



import re
import os


class VueComponent:
    """
    Parses a vue file to be understood by python
    """

    def __init__(self, location, file_name):
        self.template = ""
        self.script = ""
        self.style = ""
        self.name = file_name.replace(".vue", "")
        self.location = location
        self.file_name = file_name

        with open(os.path.join(self.location, self.file_name), "r") as f:
            self._raw = f.read()

        self.imports = self._get_imports()

    def _get_imports(self):
        """
        Gets all import statements and sends the components to the list
        """

        results = re.findall(r"import (.*)", self._raw)

        return [path for path in results]

    def _get_value_from_tag(self, tag, raw):
        """
        Gets all value inside a tag
        """

        result = re.search(f"<{tag}>(.*)</{tag}>", raw.replace("\n", ""))

        if result:
            return str(result.group(1)).strip()
        else:
            raise ValueError(f"Could parse {tag} tag.")

    def _get_script(self, raw):
        """
        Gets all value inside the script tag
        """

        script = self._get_value_from_tag("script", raw)
        result = re.search(r"export default {(.*)}", script)

        if result:
            return str(result.group(1)).strip()
        else:
            raise ValueError(f"Could parse script")

    def _replace_context(self, context):
        """
        Gets all $(key) and replaces them with the correct value.
        Raises an exception if can't find a value to the key.
        """

        replaced = self._raw
        results = re.findall(r"\$\((.*)\)", self._raw)

        for key in results:
            if key not in context:
                raise ValueError(f"Missing key {key} for component {self.name}")

            replaced = replaced.replace(f"$({key})", str(context[key]))

        return replaced

    def _load(self, raw):
        """
        Returns parsed script and template tag once raw has been replaced.
        """

        script = self._get_script(raw)
        template = self._get_value_from_tag("template", raw)
        style = self._get_value_from_tag("style", raw)

        return (script, template, style)

    def mount(self, context={}):
        """
        Replaces values and loads the component
        """

        replaced = self._replace_context(context)
        script, template, style = self._load(replaced)

        self.script = script
        self.template = template
        self.style = style

        return self

    def to_component(self, context={}):
        """
        Returns in a component form.
        """
        self.mount(context)
        return f"Vue.component('{self.name}',{{ template:'{self.template}', {self.script} }})"



import unittest

from djavue.component import VueComponent
from djavue.component_list import VueComponentList


class TestVueFile:
    def __init__(self, path):
        self.origin = "djavue/tests/src/" + path


class TestingEngine:
    def get_template(self, path):
        return TestVueFile(path)


class VueComponentListTestCase(unittest.TestCase):
    def setUp(self):
        self.test_file_path = "djavue/tests/src/"
        self.test_file_name = "component.vue"
        self.component_list = VueComponentList()
        self.component_list.engine = TestingEngine()

    def test_load(self):
        self.component_list.load("component")

        self.assertEqual(len(self.component_list.components), 2)

    def test_load_duplicate(self):
        self.component_list.load("component")
        self.component_list.load("imported")

        self.assertEqual(len(self.component_list.components), 2)

    def test_from_file(self):
        component_list = VueComponentList.from_file(
            self.test_file_path, self.test_file_name, engine=TestingEngine()
        )
        names = [c.name for c in component_list.components]

        self.assertEqual(component_list.root.name, "component")
        self.assertEqual(len(names), 2)
        self.assertIn("imported", names)
        self.assertIn("component-a", names)


class VueComponentTestCase(unittest.TestCase):
    def setUp(self):
        self.test_file_path = "djavue/tests/src/"
        self.test_file_name = "component.vue"
        self.component_list = VueComponentList.from_file(
            self.test_file_path, self.test_file_name, engine=TestingEngine()
        )
        self.component = self.component_list.root

    def test_initialization(self):
        with open(self.test_file_path + self.test_file_name, "r") as f:
            content = f.read()

        self.assertEqual(self.component.name, "component")
        self.assertEqual(self.component.location, self.test_file_path)
        self.assertEqual(self.component.file_name, self.test_file_name)
        self.assertEqual(self.component._raw, content)

    def test_import(self):
        names = [c.name for c in self.component_list.components]

        self.assertEqual(len(names), 2)
        self.assertIn("imported", names)
        self.assertIn("component-a", names)

    def test_mount(self):
        component = self.component.mount({"name": "test"})

        self.assertEqual(component.template, "<h1>{{name}}</h1>")
        self.assertEqual(component.script, 'data: () => ({        name: "test"    })')
        self.assertEqual(self.component.style, "h1{  color: 'red'}")

    def test_to_component(self):
        self.assertEqual(
            self.component.to_component({"name": "test"}),
            """Vue.component('component',{ template:'<h1>{{name}}</h1>', data: () => ({        name: "test"    }) })""",
        )



import os
import re

from djavue.component import VueComponent


class VueComponentList:
    """
    Responsible for storing and loading components.
    """

    def __init__(self):
        self.root = None
        self.engine = None
        self.root_path = ""
        self.components = []

    def _is_loaded(self, location):
        return location in [c.location for c in self.components]

    def load(self, path):
        """
        Loads a component to the list, if it's already loaded then it does nothing.
        """

        file_name = path.replace(";", "").replace('"', "")
        file_name += ".vue"

        location, file_name = os.path.split(
            str(self.engine.get_template(file_name).origin)
        )

        if self._is_loaded(location):
            return

        component = VueComponent(location, file_name)
        self.components.append(component)

        for component_path in component.imports:
            self.load(component_path)

    @staticmethod
    def from_file(path, file_name, engine=None):
        """
        Creates a new instance based on the root component location.
        """

        component_list = VueComponentList()
        component_list.root_path = path
        component_list.engine = engine

        root = VueComponent(path, file_name)
        component_list.root = root

        for component_path in root.imports:
            component_list.load(component_path)

        return component_list



import os
from typing import List

from django.template.backends.django import DjangoTemplates

from djavue.component_list import VueComponentList


class VueRenderer:
    """
    Gets template info and returns the appropriate html
    """

    def __init__(
        self,
        title: str,
        template_name: str,
        version: str = "development",
        engine: DjangoTemplates = None,
    ) -> "VueRenderer":
        self.title = title
        self.html = ""
        self.version = version

        if not engine:
            self._load_engine()
        else:
            self.engine = engine

        self.path, self.template_name = os.path.split(
            str(self.engine.get_template(template_name).origin)
        )

        self.component_list = VueComponentList.from_file(
            self.path, self.template_name, engine=self.engine
        )

    # UTILS
    def _load_engine(self) -> None:
        """
        Creates the Django Engine instance that we use to get path from file name
        """
        try:
            from django.conf import settings

            obj = settings.TEMPLATES[0].copy()
            obj["NAME"] = "a"
            obj.pop("BACKEND")
            self.engine = DjangoTemplates(params=obj).engine
        except:
            print("Error loading engine. Probably will break.")
            pass

    def _get_cdn(self) -> str:
        """
        Chooses which Vue CDN to get based on version
        """

        if self.version == "production":
            return '<script src="https://cdn.jsdelivr.net/npm/vue@2"></script>'
        else:
            return (
                '<script src="https://cdn.jsdelivr.net/npm/vue@2/dist/vue.js"></script>'
            )

    def _get_styles(self, context: object) -> str:
        root_css = self.component_list.root.mount(context).style
        components_css = [c.mount().style for c in self.component_list.components]

        return "".join([root_css] + components_css)

    # BUILDER FUNCTIONS
    def _write_header(self, context: object, style_sheets: List[str] = []) -> None:
        """
        Writes the head tag to the html
        """

        style_sheets_html: str = "".join(
            map(lambda s: f'<link href="{s}" rel="stylesheet">', style_sheets)
        )

        self.html += f"<html><head><title>{self.title}</title>{self._get_cdn()}{style_sheets_html}<style>{self._get_styles(context)}</style></head>"

    def _write_components(self, context: object) -> str:
        """
        Returns all Vue Components in component form, ready for html
        """

        return ";".join(
            map(lambda c: c.to_component(context), self.component_list.components)
        )

    def _write_body(self, context: object, scripts: List[str] = []) -> None:
        """
        Writes the body tag to the html
        """
        root = self.component_list.root.mount(context)

        scripts_html: str = "".join(
            map(lambda s: f'<script src="{s}"></script>', scripts)
        )

        self.html += f"<body><div id='root'>{root.template}</div><script>{self._write_components(context)}; new Vue({{el:'#root', {root.script} }})</script>{scripts_html}</body>"

    # RENDERINGS
    def render(
        self,
        context: object = {},
        style_sheets: List[str] = [],
        scripts: List[str] = [],
    ) -> str:
        """
        Creates the html based on the root Vue Component passed
        """

        self._write_header(context, style_sheets)
        self._write_body(context, scripts)

        return self.html


def get_vue_template(
    template_name, title="Dejavue Page", renderer=VueRenderer
) -> VueRenderer:
    """
    Loads a Vue file by it's file name and returns a VueRenderer instance ready for being rendered
    """

    return renderer(title, template_name)



from typing import List
from .base import VueRenderer


class VuetifyRenderer(VueRenderer):
    def _write_body(self, context: object, scripts: List[str] = []) -> None:
        """
        Writes the body tag to the html
        """
        root = self.component_list.root.mount(context)

        scripts_html: str = "".join(
            map(lambda s: f'<script src="{s}"></script>', scripts)
        )

        self.html += f"<body><div id='root'><v-app><v-main>{root.template}</v-main></v-app></div>{scripts_html}<script>{self._write_components(context)} new Vue({{el:'#root', vuetify: new Vuetify() {root.script} }})</script></body>"



import unittest

from djavue.renderers import VueRenderer
from djavue.tests.component import TestingEngine


class VueRendererTestCase(unittest.TestCase):
    def setUp(self):
        self.title = "New Title"
        self.renderer = VueRenderer(self.title, "component.vue", engine=TestingEngine())

    def test_write_header(self):
        self.renderer._write_header({"name": "test"})

        self.assertEqual(
            self.renderer.html,
            f"<html><head><title>{self.title}</title>{self.renderer._get_cdn()}<style>h1{{  color: 'red'}}</style></head>",
        )

    def test_write_component(self):
        context = {"name": "test_name"}
        components = self.renderer._write_components(context).split(";")

        self.assertEqual(len(components), 2)
        self.assertEqual(
            components[0],
            self.renderer.component_list.components[0].to_component(context),
        )
        self.assertEqual(
            components[1],
            self.renderer.component_list.components[1].to_component(context),
        )
